import AutoRPE.UtilsRPE.Classes.Interface as Interface
import AutoRPE.UtilsRPE.Classes.Subprogram as Subprogram
import AutoRPE.UtilsRPE.Classes.Procedure as Procedure
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.MaskCreator as MaskCreator
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.SourceManager as SourceManager
import AutoRPE.UtilsRPE.Error as Error
import re


def find_called_function(string):
    return re.match(RegexPattern.call_to_function, string.strip(), re.I).group(1)


def find_call_arguments(string):
    """Given a line with a call to a function, return different arguments, splitting by commas"""

    # Mask to avoid matching commas inside characters arguments
    dictionary_of_substitution = {}
    string = MaskCreator.mask_literal(string, dictionary_of_substitution)

    # Check contains a call
    first_parenthesis = re.search(r'\(', string)
    if not first_parenthesis:
        return []
    else:
        # Remove function name
        string = string[first_parenthesis.start():]
    # Remove first, last brackets
    string = BasicFunctions.remove_outer_brackets(string)

    # Call without arguments
    if not string.strip():
        return []

    # Index that mark if we are inside an array
    bracket_balance = 0
    for char_index, c in enumerate(string):
        if c in ["(", "["]:
            bracket_balance -= 1
        elif c in [")", "]"]:
            bracket_balance += 1
        elif c == "," and bracket_balance == 0:
            # Mark commas outside arrays, or function calls
            string = string[:char_index] + "#" + string[char_index+1:]
        else:
            continue
    # Unmask character arguments
    if dictionary_of_substitution:
        string = MaskCreator.unmask_from_dict(string, dictionary_of_substitution)
    if string.count("#"):
        return [a.strip() for a in string.split("#")]
    else:
        return [string.strip()]


def is_call_to_function(contents, procedure, vault, exclude_intrinsics=False, called_function=[]):
    from AutoRPE.UtilsRPE.NatureDeterminer import is_array
    """
    Identify if a certain string is a call to a function or it's an array.
    """
    # Check if matches the syntax of a function call
    call_to_function = re.match(RegexPattern.call_to_function, contents.strip(), re.I)
    if not call_to_function:
        return False

    # Check brackets match
    if not contents == BasicFunctions.close_brackets(contents):
        return False

    # Check if one of the argument is a slice: then is an array (matches function call syntax)
    if is_array(contents, check_sliced=True):
        return False

    function_name = find_called_function(contents)

    # Match the ret val of a function or fortran builtins (IF(.false.) will match function regex)
    if (isinstance(procedure, Subprogram.Function) and function_name == procedure.return_value.name) or \
            function_name.lower() in SourceManager.Fortran_builtins:
        return False

    # If is call to intrinsic
    if function_name.lower() in list(SourceManager.list_of_intrinsics):
        if exclude_intrinsics:
            return False
        else:
            called_function.append(function_name.lower())
            return True
    try:
        called_function.append(vault.procedures_dictionary[function_name])
    except KeyError:
        # Must be a variable
        try:
            var = vault.variables_dictionary[function_name]
            # Match variables ensuring that it is not the ret_val of a function with the same name
            if var[0].procedure.name != var[0].name:
                return False
            else:
                print("Ma che")
        except KeyError:
            # External function, if vault externals have not yet been filled
            raise Error.ProcedureNotFound

    # It is a subprogram
    return True


def get_procedure_inline(line, vault, procedure, only_subroutine=False, only_function=False):

    called_subroutines = []
    called_functions = []

    if not only_subroutine:
        called_functions = get_function(line, procedure, vault)

    # Search for calls to subprograms
    if not only_function:
        called_subroutines = get_subroutine(line, procedure, vault)

    return called_subroutines + called_functions


def get_function(string, procedure, vault):
    from AutoRPE.UtilsRPE.NatureDeterminer import is_function_declaration_statement
    if string.count("::") or is_function_declaration_statement(string):
        return []
    # This will avoid error with regex for functions that has a look ahead with exactly one space
    return _get_subprogram(re.sub(r'\s\s+', ' ', string), procedure, vault, RegexPattern.call_to_function,
                           Subprogram.Function)


def get_subroutine(string, procedure, vault):
    return _get_subprogram(string, procedure, vault, RegexPattern.call_to_subroutine, Subprogram.SubRoutine)


def _get_subprogram(string, procedure, vault, regex_pattern, procedure_type):
    import AutoRPE.UtilsRPE.Getter as Getter
    string = string.strip()
    string = MaskCreator.remove_literals(string)
    to_return = []
    external_subprogram = []
    is_unknown_external = False
    # Check is not a function or subroutine declaration, to avoid matching with functions syntax
    if re.search(RegexPattern.subprogram_declaration, string, re.I):
        return to_return

    # Avoid structures to be identified as subprogram
    purged_string = MaskCreator.mask_arrays_and_structures(string, only_structures=True)
    matches = [m for m in re.finditer(regex_pattern, purged_string, flags=re.I)]
    if matches:
        for m in matches:
            subprogram_name = m.group(1)

            # Get the actual call
            call = re.sub(r'\bcall\b', '', BasicFunctions.close_brackets(string[m.start():]), flags=re.I)

            # Check is not a variable, a builtin function or an intrinsic
            if procedure_type == Subprogram.Function:
                try:
                    if not is_call_to_function(call, procedure, vault, exclude_intrinsics=True):
                        continue
                except Error.ProcedureNotFound:
                    # An external function was hit, create an external subprogram to be returned as the error
                    is_unknown_external = True
                    subprogram = [procedure_type(subprogram_name, module=None, is_external=True)]

            if not is_unknown_external:
                # We ruled out the possibility of variables or intrinsics: must be a subprogram
                try:
                    subprogram = vault.procedures_dictionary[subprogram_name]
                except KeyError:
                    # An external subroutine was hit, create an external subprogram to be returned as the error
                    is_unknown_external = True
                    subprogram = [procedure_type(subprogram_name, module=None, is_external=True)]

            # Find called arguments
            called_arguments = find_call_arguments(call)

            if len(subprogram) > 1:
                subprogram = [s for s in subprogram if s.module.name == procedure.module.name]

            subprogram = subprogram[0]

            if not isinstance(subprogram, procedure_type):
                # Search for the correct subprogram
                if isinstance(subprogram, Interface.Interface):
                    subprogram = Getter.get_interface(called_arguments,
                                                      interface=subprogram,
                                                      block=procedure,
                                                      vault=vault)
                    # If the subprogram is not of the correct type, skip the occurrence
                    if not isinstance(subprogram, procedure_type):
                        continue
                elif isinstance(subprogram, Procedure.DerivedType):
                    continue
                else:
                    raise Error.ProcedureNotFound("The subprogram %s could not be found", subprogram_name)

            subprogram_call = create_SubprogramCall(call, subprogram, procedure, vault)
            if is_unknown_external:
                external_subprogram.append(subprogram_call)
            else:
                to_return.append(subprogram_call)
    # Raise an error returning all the unknown subprogram calls found in line
    if external_subprogram:
        raise Error.ProcedureNotFound(external_subprogram)
    return to_return


def create_SubprogramCall(call, subprogram, procedure, vault):
    # Interfaces can be called both with function name or the interface name
    subprogram_name = subprogram.name
    if subprogram.interface:
        if re.search(RegexPattern.name_occurrence % subprogram.interface.name, call):
            subprogram_name = subprogram.interface.name
    # Create the obj
    subprogram_call = SubprogramCall(subprogram_name, subprogram, procedure)
    # Line_info may not be filled yet, pass the call and fill dummy and dummy arguments
    subprogram_call.fill_dummy_arg(call)
    subprogram_call.update_actual_arguments(call=call, vault=vault)
    return subprogram_call


class SubprogramCall:
    def __init__(self, name, subprogram, block):
        self.name = name
        self.subprogram = subprogram
        self.block = block
        self.arguments = []
        self.dummy_arguments = []
        self.cast = None
        self.line_info = None
        self.index = 0

    @property
    def call(self):
        """Search for subprogram self.name on given line number, and return the call"""
        import re
        if self.line_info is None:
            raise Error.ExceptionNotManaged("Line info has not been filled yet")
        line = self.line_info.line
        # Get the correct call based on the index (in case there is more than one occurrence)
        match = re.finditer(RegexPattern.name_occurrence % self.name, line, re.I)
        start = [m.start() for m in match][self.index]
        return BasicFunctions.close_brackets(line[start:])

    @call.setter
    def call(self, updated_call):
        """Update the call by updating the proper module line"""
        import re
        if self.line_info is None:
            raise Error.ExceptionNotManaged("Line info has not been filled yet")
        module = self.block.module
        line_number = self.line_info.line_number
        line = module.lines[line_number]
        # Get the correct call based on the index (in case there is more than one occurrence)
        match = re.finditer(RegexPattern.name_occurrence % self.name, line, re.I)
        start = [m.start() for m in match][self.index]
        line = line[:start] + updated_call
        module.lines[line_number] = line


    def fill_dummy_arg(self, call=None):
        import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
        if call is None:
            call = self.call
        called_arguments = find_call_arguments(call)
        if self.subprogram.is_external:
            return
        self.dummy_arguments = []
        for index, argument in enumerate(called_arguments):
            key_arg = []
            if not self.subprogram.is_external:
                if NatureDeterminer.is_keyword_argument(argument, key_arg):
                    dummy_var = self.subprogram.get_variable_by_name(key_arg[0])
                else:
                    dummy_var = self.subprogram.get_variable_by_position(index)
                self.dummy_arguments.append(dummy_var)

    def update_actual_arguments(self, vault, call=None):
        import AutoRPE.UtilsRPE.Getter as Getter
        import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer

        if call is None:
            call = self.call
        called_arguments = find_call_arguments(call)
        self.arguments = []
        for index, argument in enumerate(called_arguments):
            key = False
            try:
                key_arg = []
                if NatureDeterminer.is_keyword_argument(argument, key_arg):
                    argument = key_arg[1]
                    var_type = Getter.get_type_of_contents(argument, procedure=self.block, vault=vault)
                else:
                    var_type = Getter.get_type_of_contents(argument, procedure=self.block, vault=vault)
            except Error.VariableNotFound:
                # An external variable has been used
                var_type = "external"

            if not var_type == "external":
                if not var_type == "char":
                    try:
                        var = Getter.get_variable(argument, procedure=self.block, vault=vault)
                    except Error.VariableNotFound:
                        var = None
                    var_type = var_type if var_type is not None else var.type
                else:
                    var = None
                arg = Subprogram.Argument(argument, key, var_type, var)
            else:
                arg = Subprogram.Argument(argument, _type="external", var=None)

            self.arguments.append(arg)

