import AutoRPE.UtilsRPE.Classes.Variable as Variable
import AutoRPE.UtilsRPE.Classes.Module as Module
import AutoRPE.UtilsRPE.Classes.Subprogram as Subprogram
import AutoRPE.UtilsRPE.Classes.Interface as Interface
import AutoRPE.UtilsRPE.Classes.Procedure as Procedure
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.MaskCreator as MaskCreator
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.AttributeParser as AttributeParser
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.Error as Error
import re


def parse_sources(file_list, vault, extension):
    # Parsing subprograms and storing info in Module structures.
    n_of_files = len(file_list)
    print("\n")
    for index, file in enumerate(file_list):
        print("\rParsing subprograms and variables %i/%i " % (index + 1, n_of_files), end="")
        module = Module.Module(file, extension)
        find_subroutines(module, vault)
        find_functions(module, vault)
        find_module_variables(module)
        vault.modules.append(module)
        # Add it after the main variables have been found, so as to add them to the variable dictionary as well
        vault.add_procedure(module.main)

    # Adding info on internal declaration of subroutines
    for procedure in vault.procedures_dictionary:
        declares = vault.procedures_dictionary[procedure][0].declares
        if declares:
            for declared_sbr in declares:
                vault.procedures_dictionary[declared_sbr][0].scope = vault.procedures_dictionary[procedure][0]

    print("\nParsing Interfaces\n")
    # Parsing the interfaces and storing in Module.
    find_interfaces(vault)

    # Fill vault variable list:
    for procedures in vault.procedures_dictionary.values():
        for proc in procedures:
            if not isinstance(proc, Interface.Interface):
                for var in proc.variables:
                    vault.variables.append(var)

    print("Parsing derived types\n")
    # Parsing the derived types
    for module in vault.modules:
        find_types(module, vault)

    # Fill blocks, and store used modules info
    for index, module in enumerate(vault.modules):
        module.create_block(vault)
        # Find used module and store
        find_used_modules(module, vault)

    # After having all the used parsed, propagate the inheritance, and add public var to available vars
    print("Expanding use statements\n")
    vault.expand_use_statement()

    # Store info of derived types in vault procedures
    for dt in vault.derived_types:
        for var in dt.members:
            vault.variables.append(var)

    # Create an has table where variables identified by name + procedure + module
    vault.fill_hash_table()

    # Add to variables info about their belongings in namelist, to be used when inserting the %sbits info
    for module in vault.modules:
        find_namelist_variable(module, vault)

    # Now that vault.derived_types is full add extension to derived types if any
    vault.expand_derived_types()

    # Store the external used data types
    vault.store_external_data_structures()

    # Stores list of function/subroutines called on each lines, stores external procedures, add level info
    fill_line_info_call(vault)

    # Fill var info on used in external
    for procedure in vault.dictionary_of_calls:
        list_of_call = vault.dictionary_of_calls[procedure]
        if list_of_call[0].subprogram.is_external:
            for call in list_of_call:
                for arg in call.arguments:
                    variable = arg.variable
                    if variable is not None:
                        variable.is_used_in_external = call.name

    # Passing to pointed objects the actual function/subroutine they points to
    process_pointers(vault)
    n_func = len([p for p in vault.procedures_dictionary.values() if isinstance(p[0], Subprogram.Function)])
    n_sbr = len([p for p in vault.procedures_dictionary.values() if isinstance(p[0], Subprogram.SubRoutine)])
    n_interfaces = len([p for p in vault.procedures_dictionary.values() if isinstance(p[0], Interface.Interface)])
    print("number of functions " + str(n_func) + "\n" +
          "number or subroutines " + str(n_sbr) + "\n" +
          "number or interfaces " + str(n_interfaces) + "\n" +
          "number or modules " + str(len(vault.modules)) + "\n" +
          "number of variables " + str(len(vault.variables)) + "\n" +
          "number or derived_types " + str(len(vault.derived_types)) + "\n")

    return vault


def get_dummy_arguments(module, index):
    # Definition: Given a line number, returns the subroutine name and the dummy arguments (in case they exist)
    # Input: integer with line number
    # Output: tuple with routine name and list of dummy arguments
    line = module.lines[index]
    # Workaround to solve the problem that appears when there's a function declaration with (kind=...)
    if line.lower().count("function "):
        if line.find("(") < line.lower().find("function "):
            line = line[line.lower().find("function "):]

    # If the subroutine declaration does not have any parenthesis, this routine does not have any argument
    if not line.count("("):
        name = line.split()[1].strip()
        dummy_arguments = []
    # If otherwise it has dummy arguments, find them
    elif line.lower().count("result") or line.lower().count("bind("):
        name = line[:line.find("(")].split()[-1]
        dummy_arguments = []
        dummy_arguments += CallManager.find_call_arguments(BasicFunctions.close_brackets(line))
    else:
        name = line[:line.find("(")].split()[-1]
        dummy_arguments = CallManager.find_call_arguments(line)
    dummy_arguments = [x for x in dummy_arguments if x.strip()]
    return name, dummy_arguments


def find_subprogram_declaration(module, _subprogram_start, _subprogram_end,
                                obj_to_find=RegexPattern.variable_declaration):
    # Definition: Given the line numbers where a routine starts and ends,
    # return list of strings containing declarations
    # TODO: manage cases in which a routine contains other functions declarations.

    stop_index = [index
                  for index, line in enumerate(module.lines[_subprogram_start:_subprogram_end + 1])
                  if re.search(RegexPattern.name_occurrence % 'contains', line, re.I)
                  or re.search(RegexPattern.name_occurrence % 'interface', line, re.I)]
    # A declaration found stop the search there
    if stop_index:
        _subprogram_end = _subprogram_start - 1 + stop_index[0]

    # Lines inside the routine
    _subprogram_lines = module.lines[_subprogram_start:_subprogram_end + 1]

    # Lines that are variable declarations
    _all_declaration = [_l.strip() for _l in _subprogram_lines if
                        re.search(obj_to_find, MaskCreator.mask_literal(_l), re.I) and not
                        (re.search(RegexPattern.general_allocation, _l, re.I) or
                         re.search(RegexPattern.string_array_initialization, _l, re.I))]

    return _all_declaration


def add_struct_members(file, structure):
    # Definition: Given the routine name, the name of the dummy arguments and the lines containing declarations,
    # this method finds the attributes of each dummy arguments and adds all the information
    # into the object variable "routines".
    # Input:
    #       -routine_name: string with the routine name
    #       -dummy_arguments: list of strings with argument names
    #       -declarations: list of strings with variable declarations
    try:
        for obj in file.objects:
            if obj.name == "main":
                main = obj
                break
        for member in structure.variables:
            main.add_variable(member)

    except UnboundLocalError:
        print("Main not found for structure %s", structure.name)


def create_subprogram(module, header, routine_name, dummy_arguments, var_declarations, used_modules,
                      subprogram_type):
    if subprogram_type == Subprogram.Function:
        # Retrieve the ret_val name
        ret_type, ret_val = find_function_type_from_header(header, var_declarations)
        if not ret_val:
            ret_val = routine_name
    elif subprogram_type == Subprogram.SubRoutine:
        ret_val = None
    else:
        raise ValueError("Header does not correspond to function or routine")

    subprogram = subprogram_type(routine_name, module)

    if len(used_modules):
        subprogram.use_only_variables = [m.lower().split("use")[1].strip() for m in used_modules]
    # Reverse approach
    dummy_arguments = [x for x in dummy_arguments]

    for declaration in var_declarations:
        m = re.search(RegexPattern.variable_declaration, declaration)
        # In this case is not a variable declaration
        if declaration.count("contiguous"):
            continue
        _attributes, variable = m.groups()
        # Check if the dummy argument is inside the declaration arguments
        # If it is, take the type, dimension, intent and optional
        _var_type, _dimension, _intent, _optional, \
        _allocatable, is_pointer, is_parameter = AttributeParser.parse_attributes(_attributes, variable)

        variable_name = AttributeParser.get_variable_name(variable)

        _v = Variable.Variable(subprogram, variable_name, _var_type, _dimension, module)
        _v.is_allocatable = _allocatable
        _v.is_parameter = is_parameter

        # If the variable name is the same of the function or is return statement
        if ret_val is not None and _v.name.lower() == ret_val.lower():
            _v.is_retval = True
            subprogram.return_value = _v
            subprogram.type = _v.type
        _v.is_pointer = is_pointer
        if variable_name in dummy_arguments:
            _v.intent = _intent
            _v.is_optional = _optional
            _v.is_dummy_argument = True
            subprogram.dummy_arguments.append(_v)
        subprogram.add_variable(_v)

    # Save the position of the argument
    found_vars = [variable.name for variable in subprogram.variables]

    # Add position to dummy arguments
    for index, dummy_argument in enumerate(dummy_arguments):
        # Check that the dummy are all correctly spelled, otherwise a variable will be duplicated
        if dummy_argument not in found_vars:
            raise Error.VariableNotFound("The variable %s is misspelled in subprogram %s declaration in module %s "
                                         % (dummy_argument, routine_name, module.name))
        dummy_argument = subprogram.get_variable_by_name(dummy_argument)
        dummy_argument.position = index
        if not dummy_argument.is_optional:
            subprogram.mandatory_da.append(dummy_argument)
        for m_i, m_da in enumerate(subprogram.mandatory_da):
            m_da.mandatory_position = m_i
    # Order dummy_arguments by position
    subprogram.dummy_arguments.sort(key=lambda x: x.position)
    if subprogram_type == Subprogram.Function:
        # In case of function  declaration like INTEGER FUNCTION name() without return variable declaration
        if not [var for var in subprogram.variables if var.is_retval]:
            # Add the return value to functions variables
            _v = Variable.Variable(subprogram, ret_val, ret_type, 0, module)
            subprogram.add_variable(_v)
            _v.is_retval = True
            # Specify the type of the function
            subprogram.type = ret_type
            subprogram.return_value = _v
    # Save object
    return subprogram


def find_subroutines(module, vault):
    #  Looks into the sourcefile for routine declarations, storing
    #  name, type, dimensions and optionality of the arguments.

    for _index, line in enumerate(module.lines):
        # Look if routine starts with (recursive) subroutine
        sbr_name = re.search(RegexPattern.subroutine_declaration, line, re.IGNORECASE)
        if sbr_name:
            sbr_name = sbr_name.group(1)
            subroutine_start = _index
            # Look for end subroutine statements
            try:
                subroutine_end = [index for index, l in enumerate(module.lines)
                                  if re.search(r'end\s+subroutine\s+%s\b' % sbr_name, l, re.I)][0]
            except IndexError:
                raise Error.SubprogramNotFound(
                    'The subprogram "end subroutine {subname}" statement from the module "{module}" has not been found.'.format(
                        subname=sbr_name, module=module.name))

            # Find name and dummy arguments
            subroutine_name, dummy_arguments = get_dummy_arguments(module, _index)
            # Find declarations
            var_declarations = find_subprogram_declaration(module, subroutine_start, subroutine_end,
                                                           obj_to_find=RegexPattern.variable_declaration)
            used_modules = find_subprogram_declaration(module, subroutine_start, subroutine_end,
                                                       obj_to_find=RegexPattern.use_statement)
            header = module.lines[subroutine_start]
            # Create subroutine objects, with variables and their attribute. Then add it to the module
            subroutine = create_subprogram(module, header, subroutine_name, dummy_arguments, var_declarations,
                                           used_modules, subprogram_type=Subprogram.SubRoutine)
            # Check if subroutines are defined inside this scope
            subroutine_lines = module.lines[subroutine_start + 1:subroutine_end - 1]
            internal_subprogram = [re.search(RegexPattern.subprogram_declaration, l, re.I) for l in subroutine_lines
                                   if re.search(RegexPattern.subprogram_declaration, l, re.I)]
            if internal_subprogram:
                subprogram_names = [i_s.group(2) for i_s in internal_subprogram]
                subroutine.declares = subprogram_names

            vault.add_procedure(subroutine)


def add_external_subprogram(unknown_calls, vault):
    for subprogram_call in unknown_calls.args[0]:
        try:
            vault.procedures_dictionary[subprogram_call.subprogram.name]
        except KeyError:
            # Stores the external function in the database
            vault.procedures_dictionary[subprogram_call.subprogram.name] = [subprogram_call.subprogram]
            for arg in subprogram_call.arguments:
                if arg.variable is not None:
                    # When producing the final list of variable we will have to exclude these variables
                    arg.variable.is_used_in_external = subprogram_call.name


def fill_line_info_call(vault):
    n_mod = len(vault.modules)
    for index, module in enumerate(vault.modules):
        print("\rFill info line module %i/%i  %s" % (index + 1, n_mod, module.name), end="")
        for line_index, line in enumerate(module.lines):
            # Remove if condition
            line = module.line_info[line_index].unconditional_line
            if_condition = module.line_info[line_index].if_condition
            block = module.line_info[line_index].block
            for line_part in line, if_condition:
                # Otherwise all function call ends up duplicated
                if not line_part.strip():
                    continue
                # Search for functions
                try:
                    called_functions = CallManager.get_function(line_part, block, vault)
                except Error.ProcedureNotFound as unknown_subroutines_calls:
                    # If unknown function was found add it to vault
                    add_external_subprogram(unknown_subroutines_calls, vault)
                    # Store this function in the array
                    called_functions = CallManager.get_function(line_part, block, vault)
                # Search for subroutines call ( after function, unknown function could be used as sbr argument)
                try:
                    called_subroutines = CallManager.get_subroutine(line_part, block, vault)
                except Error.ProcedureNotFound as unknown_subroutines_calls:
                    # If unknown subroutine was found add it to vault
                    add_external_subprogram(unknown_subroutines_calls, vault)
                    called_subroutines = CallManager.get_subroutine(line_part, block, vault)
                # Add the information to line info and fill vault dictionary of calls
                subprogram_call = called_functions + called_subroutines
                for i_s_call, s_call in enumerate(subprogram_call):
                    # Name will be subprogram name + module name, to avoid confusion among subprograms with same name
                    if not s_call.subprogram.is_external:
                        name = s_call.subprogram.module.name + "_mp_" + s_call.subprogram.name
                    else:
                        name = s_call.subprogram.name
                    s_call.line_info = module.line_info[line_index]
                    # Store the index of the call wrt the total number of this function's call on the line
                    s_call.index = len([c for c in subprogram_call[:i_s_call] if c.name == s_call.name])
                    try:
                        vault.dictionary_of_calls[name] += [s_call]
                    except KeyError:
                        vault.dictionary_of_calls[name] = [s_call]
    # Add level info to subprogram calls
    _fill_subprogram_level(vault)
    print("\n")


def _fill_subprogram_level(vault):
    # Recursive function to compute the level of the function in the pipeline
    def compute_procedure_level(p, v):
        # If the subprogram is not calling any function its level is 0
        if not p.internal_call:
            p.level = 0
            return 0
        # If there are internal calls, compute the pipeline level
        set_depth = [-1]*len(p.internal_call)
        for i, p_name in enumerate(p.internal_call):
            internal_procedure = v.procedures_dictionary[p_name][0]
            # The level of the internal call has already been computed
            if internal_procedure.level != -1:
                set_depth[i] = internal_procedure.level
            else:
                # We hit a recursive functions
                if p.name in internal_procedure.internal_call:
                    internal_procedure.level = 1
                else:
                    # Recursively compute the internal function level
                    internal_procedure.level = compute_procedure_level(internal_procedure, v)
                set_depth[i] = internal_procedure.level
        # Return the biggest among the level. the +1 accounts for this function extra-layer
        return max(set_depth) + 1

    # Store subprograms called by each subprogram
    for call_name in vault.dictionary_of_calls:
        # Using the dictionary of calls, we look where the call happened
        call = vault.dictionary_of_calls[call_name]
        for c in call:
            # Retrieve the object describing the scope
            scope = vault.procedures_dictionary[c.block.name]
            # To avoid this we should store block name with subprogram_name_mp_module_name
            if len(scope) != 1:
                Error.ExceptionNotManaged("One function expected, got two")
            # Retrieve the actual name, in case it was a call to an interface
            name = c.subprogram.name
            # Check the call uses dummy arguments of real type (we do not care if the calls are not entangled)
            used_da = [da for da in scope[0].dummy_arguments if da.type in VariablePrecision.real_id if
                       da in [a.variable for a in c.arguments]]
            if name not in scope[0].internal_call and used_da:
                scope[0].internal_call.extend([name])

    # For each function in the dictionary, compute the "level" in the pipeline
    for procedure_name in vault.procedures_dictionary:
        if procedure_name == 'main':
            continue
        procedure = vault.procedures_dictionary[procedure_name][0]
        if isinstance(procedure, Subprogram.SubRoutine) or isinstance(procedure, Subprogram.Function):
            if procedure.level == -1:
                procedure.level = compute_procedure_level(procedure, vault)


def find_types(module, vault):
    # Looks into the module lines for type declaration
    in_type_declaration = False
    derived_types = []
    type_name = None
    variable_counter = None
    for _index, line in enumerate(module.lines):
        # Avoid match with line with 'TYPE' in string
        line = MaskCreator.mask_literal(line)
        # Look if routine starts with subroutine
        if not in_type_declaration:
            # Check if a type has been declared
            m = re.search(RegexPattern.type_declaration, line, re.I)
            if m:
                in_type_declaration = True
                variable_counter = 0
                type_name = m.group(3)
                extended_type = None
                # Case in which a type is extending another type, search the extended type name
                extend_statement = re.search(RegexPattern.name_occurrence % "extends", line, re.I)
                if extend_statement:
                    extend_statement = BasicFunctions.close_brackets(line[extend_statement.start():])
                    extended_type = CallManager.find_call_arguments(extend_statement)[0]
                # Create the structure
                derived_type = Procedure.DerivedType(type_name, module, extended_type)
                derived_types.append(derived_type)
        else:
            # Check if we are exiting a type declaration
            m = re.search(RegexPattern.end_type_statement, line, re.IGNORECASE)
            if m:
                in_type_declaration = False
            else:
                m = re.search(RegexPattern.variable_declaration, line)
                if m:
                    attributes, variables = m.groups()
                    var_type, dimension, intent, optional, \
                    allocatable, is_pointer, is_parameter = AttributeParser.parse_attributes(attributes, variables)

                    variable_name = AttributeParser.get_variable_name(variables)

                    v = Variable.Variable(derived_type, variable_name, var_type, dimension, module)
                    v.position = variable_counter
                    v.is_dummy_argument = True
                    v.is_allocatable = allocatable
                    v.is_parameter = is_parameter
                    v.is_pointer = is_pointer
                    v.is_member_of = type_name
                    derived_type.add_member(v)
                    variable_counter += 1
    # Now that also the variables have been added to data structure we can fill the dictionary
    for dt in derived_types:
        vault.derived_types.append(dt)
        vault.add_procedure(dt)


def find_module_variables(module):
    line_to_check = True

    for line in module.lines:
        if line_to_check:
            # Check if we are entering a type declaration zone
            is_type_declaration = re.search(RegexPattern.type_declaration, line, re.IGNORECASE)
            if is_type_declaration or NatureDeterminer.is_function_declaration_statement(line):
                line_to_check = False
                continue
        else:
            # Check if we are exiting a type declaration zone
            is_end_type_declaration = re.search(RegexPattern.end_type_statement, line, re.IGNORECASE)
            if is_end_type_declaration or NatureDeterminer.is_end_function_statement(line):
                line_to_check = True
            # Even if we are inside or exiting a type declaration zone, we have to jump to next line
            continue

        if line.lower().count("contains"):
            break
        if line.lower().count("end program"):
            break
        match = re.search(RegexPattern.variable_declaration, line)
        if match:
            attributes, variables = match.groups()
            var_type, dimension, intent, optional, \
            allocatable, is_pointer, is_parameter = AttributeParser.parse_attributes(attributes, variables)
            variable_name = AttributeParser.get_variable_name(variables)
            # It means is a variable and not a public subprogram
            if var_type:
                v = Variable.Variable(module.main, variable_name, var_type, dimension, module)
                v.is_allocatable = allocatable
                v.is_pointer = is_pointer
                v.is_parameter = is_parameter
                module.main.add_variable(v)
                module.main.add_accessible_var(v)
                module.variables.append(v)


def find_namelist_variable(module, vault):
    if module.get_text().lower().count('namelist'):
        for index, line in enumerate(module.lines):
            if line.lower().strip().find("namelist") == 0:
                # Retrieve all variables in the namelist and set the module and name of the namelist
                namelist_variables = [var.strip() for var in line.split("/")[2].split(",")]
                namelist_name = re.search(r'namelist\s*/(.*)/', line, re.I).group(1)
                for nv in namelist_variables:
                    var = vault.get_variable_by_name(nv, module.line_info[index].block)
                    var.appears_in_namelist = [module.module_name, namelist_name]


def find_functions(module, vault):
    # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    for _index, line in enumerate(module.lines):
        if NatureDeterminer.is_function_declaration_statement(line):
            function_start = _index
            # Look for end subroutine statements
            function_end = BasicFunctions.find_next_end(module.lines, _index, "end function")

            # Find name and dummy arguments
            function_name, dummy_arguments = get_dummy_arguments(module, _index)
            # Find declarations
            variables_declarations = find_subprogram_declaration(module, function_start, function_end, obj_to_find="::")
            used_modules = find_subprogram_declaration(module, function_start, function_end, obj_to_find="^use\s")
            # Parse everything and store
            header = module.lines[function_start]
            function = create_subprogram(module, header, function_name, dummy_arguments, variables_declarations,
                                         used_modules, subprogram_type=Subprogram.Function)
            vault.add_procedure(function)


def find_interfaces(vault):
    for module in vault.modules:
        # Definition: Looks into the sourcefile for interface declarations, storing the name.
        for _index, _line in enumerate(module.lines):
            if _line.strip().lower().find("interface") == 0 or _line.strip().lower().find("abstract interface") == 0:
                interface_start = _index
                interface_end = BasicFunctions.find_next_end(module.lines, _index, "end interface")
                try:
                    # Check if it is abstract or not
                    if _line.strip().lower().find("abstract interface") == 0:
                        interface_name = None
                        is_abstract = True
                    else:
                        interface_name = _line.split()[1].strip()
                        is_abstract = False
                except IndexError:
                    # interface to an external procedure or a procedure used in a procedure with the DEFERRED attribute
                    continue
                interfaced_procedure = []
                for interface_line in module.lines[interface_start:interface_end + 1]:
                    if interface_line.strip().lower().find("module procedure") == 0:
                        line_procedure = interface_line.replace("module procedure", "")
                        line_procedure = line_procedure.replace("MODULE PROCEDURE", "")
                        line_procedure = [_r.strip() for _r in line_procedure.split(",")]
                        interfaced_procedure += line_procedure
                    elif interface_line.strip().lower().strip().find("subroutine") == 0 \
                            or interface_line.strip().lower().strip().find("function") == 0:
                        subroutine_name = interface_line[:interface_line.find("(")].split()[1]
                        interfaced_procedure += [subroutine_name]

                # Create a name for the abstract interface based on the procedures names
                if interface_name is None:
                    interface_name = "abstract_interface_" + "_".join(interfaced_procedure).replace(" ", "_")

                interface = Interface.Interface(interface_name, module, is_abstract)
                for ip in interfaced_procedure:
                    interface.add_procedure(vault.procedures_dictionary[ip][0])
                interface.generate_key()
                vault.add_procedure(interface)
                vault.interfaces_dictionary[interface.name] = interface

    # Once the dictionary of interface is full, add the interface property "has mixed interface"
    for interface in vault.interfaces_dictionary.values():
        for subprograms in interface.agnostic_key['all'].values():
            # More than one subprogram correspond to the same agnostic key: is a sp/dp interface
            if len(subprograms) > 1:
                for s in subprograms:
                    try:
                        s = vault.procedures_dictionary[s][0]
                    except KeyError:
                        raise KeyError
                    s.has_mixed_interface = True


def find_used_modules(module, vault):
    for index, line in enumerate(module.lines):
        line = line.strip().lower()
        # First line will module name, which we want to skip
        if line.lower().find("module ") == 0 or line.lower().find("program ") == 0:
            continue
        # No need to parse the entire file
        if not line.find("use ") == 0:
            break

        module_name = line.split()[1]

        use_only = []
        if line.count(","):
            use_only = line.split(":")[1].lower().replace("only", "")
            use_only = use_only.lower().replace(":", "").strip()
            use_only = [u.strip() for u in use_only.split(",")]
        if module_name.count(","):
            # Remove the only information
            module_name = module_name.split(",")[0].strip()

        # Sometimes use statement are repeated
        # Add to the module the used modules
        used_module = vault[module_name]
        # If the module is None it means it is an external import
        if used_module is not None and used_module not in module.use:
            if use_only:
                for var in use_only:
                    try:
                        var = vault[module_name].get_variable(var, 'main')
                    except Error.VariableNotFound:
                        # It means a subprogram has been imported
                        continue
                    module.use_only.append(var)
            else:
                module.use.append(used_module)
        # Look for ?
        pattern = "(\w+) *=> *(\w+)"
        matches = re.findall(pattern, line)
        if matches:
            for match in matches:
                pointer, target = match
                target_var = vault.module.use[-1].main().get_variable_by_name(target)
                new_var = Variable.Variable(module.line_info[index].block, pointer, target_var.type,
                                                   target_var.dimension)
                module.line_info[index].block.add_variable(new_var)


def find_function_type_from_header(header, var_declaration):
    # Get the function name
    fun_name = header.lower().split('function')[1]
    fun_name = fun_name.split("(")[0]
    fun_name = fun_name.replace("(", "")
    fun_name = fun_name.strip()

    # Check if a return val is declared
    if header.lower().count("result"):
        ret_val = header.lower().split('result')[1]
        ret_val = ret_val.replace('results', '')
        ret_val = ret_val.replace('(', '')
        ret_val = ret_val.replace(')', '').strip()
    else:
        ret_val = fun_name

    # Check if the return_val typ is specified among the function variables
    declaration_of_retval = [v.split("::")[1].lower().strip() for v in var_declaration
                             if v.split("::")[1].lower().strip() in [fun_name, ret_val]]
    if declaration_of_retval:
        return None, declaration_of_retval[0]
    # The type is specified in front of the function
    pattern = "(.*)function"
    match = re.search(pattern, header.strip(), re.IGNORECASE)
    fun_type = remove_brackets_content(match.groups(0)[0].strip().lower())
    if fun_type == "real":
        real_declaration = match.group(1)
        nt = re.match(RegexPattern.precision_of_real_declaration, real_declaration, re.I)
        if nt is not None:
            try:
                fun_type = VariablePrecision.lookup_table[nt.group(3)]
            except KeyError:
                raise Error.ExceptionNotManaged("Pleas add this type specifier into VariablePrecision.lookup_table: %s"
                                                % nt.group(3))
        else:
            print("Type specifier not found in real function " + header + " declaration")
            return VariablePrecision.real_id['wp']
    if match.group(0).count('rpe_var'):
        return 'rpe_var', fun_name
    else:
        return fun_type, fun_name


def remove_brackets_content(string):
    op = [i for i, x in enumerate(string) if x == "("]
    cp = [i for i, x in enumerate(string) if x == ")"]
    _new_line = []
    for n in range(len(string)):
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if len([x for x in op if x <= n]) == len([x for x in cp if x < n]):
            _new_line += [string[n]]
    return "".join(_new_line)


def process_pointers(vault):
    for var in vault.variables:
        if var.is_pointer:
            if var.is_pointer.count("type"):
                pointer_name = var.is_pointer.replace("type", "")
                pointer_name = BasicFunctions.remove_outer_brackets(pointer_name)
                search_array = vault.derived_types
            elif var.is_pointer.count("class"):
                # Same of type, but for compiler reason they have to be declared as classes
                pointer_name = var.is_pointer.replace("class", "")
                pointer_name = BasicFunctions.remove_outer_brackets(pointer_name)
                search_array = vault.derived_types
            elif var.type == "procedure":
                pointer_name = var.is_pointer
                search_array = []
                for val in vault.procedures_dictionary.values():
                    search_array += val
            else:
                # It is a pointer to real or a pointer to integer
                continue
            pointed_obj = [obj for obj in search_array if obj.name == pointer_name]
            if len(pointed_obj) > 1:
                raise Error.DuplicatedObject
            else:
                var.is_pointer = pointed_obj
            continue
