import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.MaskCreator as MaskCreator
import AutoRPE.UtilsRPE.CallManager as CallManager
import re


def clean_sources(input_file_list):
    """  Remove comments, empty lines and old style declaration"""
    n_of_file = len(input_file_list)
    for index, file in enumerate(input_file_list):
        print("\rFormatting file %i/%i  %20s" % (index + 1, n_of_file, file.filename), end="")
        clean(file)


def clean(file):
    """Format file, so it can be parsed to generate a vault"""
    # Remove all comments
    remove_all_comments(file)

    # Remove \n char (can be found inside strings) to avoid problem with split lines function
    remove_newline_character(file)
    # Merges commands split on multiples lines
    merge_lines(file)

    # Multi command lines are split into multiple lines
    split_lines(file)

    # Split lines with more than one declaration
    split_declarations(file)

    # Remove parameter assignation F77 style
    _remove_F77_parameter_declaration(file)

    # Substitutes all multiple spaces with single spaces ( for some regex to work)
    remove_multiple_space(file)

    # Deletes all the empty lines
    remove_empty_lines(file)

    # Put text back together
    file.update_module_lines()


def split_lines(file):
    """Split into multiple lines multiple-command lines"""
    for index, line in enumerate(file.lines):
        if line.count(";"):
            n_of_spaces = BasicFunctions.count_initial_spaces(line)
            # Create a string where literals have been substituted with #
            mask = MaskCreator.mask_literal(line)
            # Substitutes semicolons with \n command
            match = re.finditer(r";", mask)
            semicolons = [m.start() for m in match]
            # Always substitute from the last match, to avoid changes in indexing
            for i_sem in semicolons[::-1]:
                line = line[:i_sem] + "\n" + " " * n_of_spaces + line[i_sem + 1:].strip()
            file.lines[index] = line
    # This will take care of translating \n command into different lines
    file.update_module_lines()


def remove_multiple_space(file):
    """Remove multiple spaces inside string, keeping indentation"""
    for index, line in enumerate(file.lines):
        dict_literal_sub = {}
        mask_line = MaskCreator.mask_literal(line, dict_literal_sub)
        n_of_spaces = BasicFunctions.count_initial_spaces(mask_line)
        mask_line = re.sub(r'\s\s+', ' ', mask_line)
        if dict_literal_sub:
            file.lines[index] = MaskCreator.unmask_from_dict(mask_line, dict_literal_sub)
        else:
            file.lines[index] = mask_line
        file.lines[index] = " " * n_of_spaces + file.lines[index]


def split_declarations(file):
    import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
    # Split declarations of variables putting a single variable in each line
    for _index, _line in enumerate(file.lines):
        if _line.count("::") and not re.search(RegexPattern.string_array_initialization, _line) and not \
                re.search(RegexPattern.general_allocation, _line, re.I):
            attributes, arguments = _line.split("::")
            arguments = CallManager.find_call_arguments("c(%s)" % arguments)
            if len(arguments) > 1:
                # Put a declaration for each line
                file.lines[_index] = "".join(["%s :: %s\n" % (attributes, arg) for arg in arguments])

    file.update_module_lines()


def remove_all_comments(file):
    # Remove comments from a file
    for _index, _line in enumerate(file.lines):
        file.lines[_index] = BasicFunctions.remove_comments(_line)


def _remove_F77_parameter_declaration(file):
    # If parameter declaration is made in Fortran 77 style
    pattern = '^\s*PARAMETER\s*\(\s*[\w+=\s*]'

    for line_index, line in enumerate(file.lines):

        # Search for lines with the 'PARAMETER' statement
        match = re.match(pattern, line, re.I)
        if match:
            param_line = line.replace(")", "")

            # Store parameter names
            parameters_name = [parameter.split("=")[0].strip() for parameter in
                               param_line.split("(")[1].split(",")]

            # Store parameter values
            parameters_val = [parameter.split("=")[1].strip() for parameter in
                              param_line.split("(")[1].split(",")]

            for parameter_name, parameter_value in zip(parameters_name, parameters_val):

                # Search for each parameter declaration backward in the file
                for dec_line_index, dec_line in enumerate(reversed((file.lines[:line_index]))):
                    parameter_pattern = '.*::\s*' + parameter_name + '$'
                    param_declaration = re.match(parameter_pattern, dec_line, re.I)
                    if (line_index - dec_line_index - 1) == 0:
                        raise AssertionError("declaration of parameter %s not found in file %s" % (
                            parameter_name, file.filename))
                    # Put back in place the parameter attribute and assign the parameter vale
                    if param_declaration:
                        # Add param statement
                        piece_to_replace = dec_line.split('::')[0]
                        replacement = "%s, parameter " % piece_to_replace
                        dec_line = dec_line.replace(piece_to_replace, replacement)

                        # Add param assignation
                        dec_line = dec_line + "=" + parameter_value

                        # Substitute the line
                        file.lines[line_index - dec_line_index - 1] = dec_line

                        break

            # Delete parameter declaration
            file.lines[line_index] = ""


def merge_lines(file):
    """Given a file, puts back on the same line commands split with a &"""
    # List split lines index
    split_index = [i for i, l in enumerate(file.lines) if l.count("&")]
    for index in split_index:
        # If is an openmp directive, skip
        omp_dir = re.search(r"^\s*#", file.lines[index])
        if omp_dir:
            continue
        # Merge the split line on first line, nullify lines that have been merged in the first one
        BasicFunctions.merge_line_in_file(index, file)


def remove_newline_character(file):
    """Remove \n character. Output will be on single line, but avoid mess up with rebuild of modules"""
    newline_index = [i for i, l in enumerate(file.lines) if l.count("\\n")]
    for index in newline_index:
        file.lines[index] = file.lines[index].replace("\\n", "")


def remove_empty_lines(file):
    """
      Remove empty strings from list of strings
    """
    file.lines = [_l for _l in file.lines if _l.strip()]
