import signal
import time
from AutoRPE.UtilsWorkflow.Communicator import SSH


class Counter:
    def __init__(self):
        self.count = 0

    def up(self, how_much=1):
        self.count += how_much
        return self.count - 1


class GracefulKiller:
    """
    This class is used to allow the user to stop an analysis at any point creating a pause_checkpoint.pkl which
    can be used in the future to restart the analysis from the same point.

    It uses the signal library to catch signals.
    """
    kill_now = False

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)
        self.filename = None

    def exit_gracefully(self, signum, frame):
        print("Exit has been activated, will write the checkpoint as soon as possible and exit.")
        self.kill_now = True


class BinaryTreeSearch:
    def __init__(self, remote_hostname, remote_user, remote_scratch,
                 local_folder, job_template, vault,
                 original_precision_level, reduced_precision_level, accuracy_test, max_running_jobs,
                 experiment_name="BinaryTreeSearch", checkpoint_name=None):

        self.experiment_name = experiment_name
        # Remote connection related parameters
        self.remote_hostname = remote_hostname
        self.remote_user = remote_user
        self.remote_scratch = remote_scratch
        self.max_running_jobs = max_running_jobs
        self.local_folder = local_folder
        # Starting SSH connection
        self.communicator = SSH(user=self.remote_user, host=self.remote_hostname, remote_scratch=self.remote_scratch)

        # Parameters of the analysis
        self.original_precision_level = original_precision_level
        self.reduced_precision_level = reduced_precision_level

        # Experiment related parameter
        self.accuracy_test = accuracy_test
        self.job_template = job_template
        self.vault = vault
        self.analysis_variables = [v for v in self.vault.variables if v.id is not None and v.is_used][:]

        # In case of starting from a checkpoint
        self.checkpoint_name = checkpoint_name

        # Initialize an empty experiment driver
        self.binary_tree = None

        # Used to give an incremental ID to jobs  (for visualization)
        self.counter = Counter()

    def setup(self, id_reduced_precision, forced_ids, id_banned_var):
        """
        This method sets up the experiment driver in order to start with the experiment.
        """
        from AutoRPE.UtilsWorkflow.Job import Job
        from AutoRPE.UtilsWorkflow.BinaryTree import BinaryTree
        print("Starting analysis with %i variables" % len(id_reduced_precision))
        # Create the starting job or recover it from a checkpoint
        if self.checkpoint_name is None:
            # Creating the first set that will include all the variables
            root = Job(id_reduced_precision=[ids for ids in id_reduced_precision if ids not in id_banned_var],
                       forced_ids=[ids for ids in forced_ids if ids not in id_banned_var],
                       analysis_variables=self.analysis_variables,
                       banned_variables=id_banned_var,
                       reduced_precision_level=self.reduced_precision_level,
                       communicator=self.communicator,
                       vault=self.vault,
                       template=self.job_template,
                       local_folder=self.local_folder,
                       result_filename=self.accuracy_test.filename_to_evaluate,
                       counter=self.counter)
            # Defining it's level to 0
            root.level = 0
            # Initialize the Tree with the root job
            binary_tree = BinaryTree(root, self.accuracy_test)
        else:
            # Loading from checkpoint
            print("Resuming from %s" % self.checkpoint_name)
            binary_tree = self.recover_checkpoint()

        # Save the experiment driver in the class object
        self.binary_tree = binary_tree

    def initial_check(self, forced_id=[], banned_id=[]):
        from AutoRPE.UtilsWorkflow.Job import Job
        import time
        # Make a original-precision run to be sure that the accuracy test works as expected
        for var in self.analysis_variables:
            var.precision = self.original_precision_level

        check_job = Job(id_reduced_precision=[],
                        forced_ids=forced_id,
                        analysis_variables=self.analysis_variables,
                        banned_variables=banned_id,
                        reduced_precision_level=self.reduced_precision_level,
                        communicator=self.communicator,
                        vault=self.vault,
                        template=self.job_template,
                        result_filename=self.accuracy_test.filename_to_evaluate,
                        local_folder=self.local_folder,
                        counter=0
                        )
        # Submit job and wait until it finishes
        check_job.run_job()
        while check_job.remote_status != "COMPLETED" and check_job.remote_status != "FAILED":
            time.sleep(5)

        # Check that the job ran successfully
        if self.accuracy_test.evaluate_success(remote_rundir=check_job.remote_rundir):
            print("Basic test was successful, continuing with the analysis")
        else:
            raise AssertionError("Basic simulation to check that the accuracy test works properly didn't pass")

    def root_job_succeeded(self):
        """Return true once the root node succeeded"""
        # In case the root of the tree fails, raise an error
        if self.binary_tree.root.status == "FAILED":
            raise AssertionError("Whole set job ended with a failed result. Should check what happened.")
        # The status of the root will be SUSPENDED until the entire tree has been checked
        return self.binary_tree.root.status == "SUCCESS"

    def run(self):
        # Pass the communicator to the accuracy test, so it can download the data
        self.accuracy_test.communicator = self.communicator
        # Before starting running the analysis, it is convenient to perform an initial check to assert that
        # setting all the analysis variables in double precision it passes the test.
        # Otherwise it would mean that we are trying to do something impossible.
        self.initial_check()

        # First run the cluster analysis
        cluster_var_root = self.run_analysis(id_reduced_precision=self.get_cluster_var(),
                                             id_forced_var=[],
                                             id_banned_var=[])

        local_var_root = self.run_analysis(id_reduced_precision=self.get_local_var(),
                                           id_forced_var=[],
                                           id_banned_var=[])

        global_var_root = self.run_analysis(id_reduced_precision=self.get_global_var(),
                                            id_forced_var=[],
                                            id_banned_var=[])

        local_cluster_var_root = self.run_analysis(id_reduced_precision=local_var_root.id_reduced_precision[:],
                                                   id_forced_var=cluster_var_root.id_reduced_precision[:],
                                                   id_banned_var=[])

        local_cluster_global_root = self.run_analysis(id_reduced_precision=local_cluster_var_root.id_reduced_precision[:],
                                                      id_forced_var=cluster_var_root.id_reduced_precision[:] +
                                                                    global_var_root.id_reduced_precision[:],
                                                      id_banned_var=[])

        for var_id in local_cluster_global_root.id_reduced_precision + local_cluster_global_root.forced_ids:
            self.vault.get_variable_by_id(var_id).precision = self.reduced_precision_level
        self.print_root_configuration(local_cluster_global_root)

        # Saving a last checkpoint
        self.binary_tree.checkpoint(iteration=1000, filename="end_checkpoint.pkl")

        # Probably not working right now
        # create_explorer(experiment=experiment, vault=self.vault)
        # Return success
        return True

    def recover_checkpoint(self):
        import pickle as pkl
        binary_tree, iteration = pkl.load(open(self.checkpoint_name, "rb"))
        all_jobs = binary_tree.all() + binary_tree.disinherited
        for job in all_jobs:
            job.communicator = self.communicator
            job.vault = self.vault
            job.template = self.job_template
            job.local_folder = self.local_folder
        job_with_no_parent = [job for job in all_jobs if job.parent is None and job.level == 0]
        root = job_with_no_parent[0]
        root.vault = self.vault
        return binary_tree

    def print_root_configuration(self, root):
        import calendar
        ts = calendar.timegm(time.gmtime())
        filename = self.experiment_name + "_root_configuration_" + str(ts)
        # Save list of variables into a txt file
        # Sort and keep only unique values
        line_format = "%5i # %20s %25s %20s %5s\n"
        with open(filename, "w") as output_file:
            # Add header
            output_file.write("   ID " + " " * 5 +
                              "#" +
                              " " * 13 +
                              "variable name" + " " * 11 +
                              "subprogram name" + " " * 9 +
                              "module name" + " " * 3 +
                              "variable precision" + "\n"
                              + 100 * "#" + "\n")
            # Put the var in ascending id order
            root.analysis_variables.sort(key=lambda x: x.id)
            # Print the variables to file
            for var in root.analysis_variables:
                output_file.write(line_format %
                                  (var.id, var.name, var.procedure.name, var.procedure.module.name, var.precision))

    def get_cluster_var(self):
        # The variables belonging to cluster have a cluster id assigned
        return [v.id for v in self.analysis_variables if v.cluster_id != -1]

    def get_global_var(self):
        # Global independent variables belong to main and are not assigned a cluster id
        return [v.id for v in self.analysis_variables if v.cluster_id == -1 and v.procedure.name == 'main']

    def get_local_var(self):
        # Variables not in clusters have assigned id -1
        return [v.id for v in self.analysis_variables if v.cluster_id == -1 and not v.procedure.name == 'main']

    def run_analysis(self, id_reduced_precision, id_forced_var, id_banned_var):
        from AutoRPE.UtilsRPE.Error import ExceptionNotManaged
        self.setup(id_reduced_precision, id_forced_var, id_banned_var)

        # Use a signal catcher to handle experiment pause
        killer = GracefulKiller()
        change_counter = 0
        # Run the whole analysis until the root job succeed
        while not self.root_job_succeeded():
            # Loop through the jobs in the queue and take actions if required
            step_change = self.binary_tree.step()

            # In case there hasn't been any change, this prevents the script of polling SLURM too much
            if step_change == 0:
                time.sleep(10.0)
                # pass
            else:
                # Check how many jobs did change status in order to save experiment checkpoints from time to time
                change_counter += step_change
                if change_counter > 10:
                    # self.binary_tree.checkpoint()
                    change_counter = 0

            # If there was a signal during this loop:
            if killer.kill_now:
                self.binary_tree.checkpoint(filename="paused_checkpoint.pkl")
                print("Exiting...")
                exit(0)
        root = self.binary_tree.root
        # Now set the reduced precision to those variables that proved safe
        for v_id in root.id_reduced_precision:
            root.vault.get_variable_by_id(v_id).precision = self.reduced_precision_level

        return self.binary_tree.root
