# This file contains the definition of a Binary Search Job, which its an extension of a Job class.
# The extension includes several methods that are useful to do a precision analysis using a binary search algorithm.
from AutoRPE.UtilsWorkflow.RemoteManager import RemoteManager
import AutoRPE.UtilsRPE.Error as Error


# Extend it
class Job(RemoteManager):
    def __init__(self,
                 id_reduced_precision,
                 forced_ids,
                 banned_variables,
                 analysis_variables,
                 reduced_precision_level,
                 communicator,
                 vault,
                 template,
                 local_folder,
                 result_filename,
                 counter
                 ):
        super().__init__(id_reduced_precision,
                         forced_ids,
                         reduced_precision_level,
                         analysis_variables,
                         communicator,
                         vault,
                         template,
                         result_filename,
                         counter)
        # This class extends the classical job putting some extra variables that define the relations
        self.parent = None
        self.child = []
        self.child_number = None
        self.reshuffle = False
        self.level = None
        self.cluster_divided = False

        # Other things that need to be saved
        self.banned_variables = banned_variables[:]
        self.local_folder = local_folder
        self.max_retrials = 1
        self.retrials = 0
        self.counter = counter

    def print_info_job(self, status):
        print("Job %s %s %s before computing fitness" % (self.incremental_id, self.hash, status))

    def ancestors(self):
        to_return = []
        cv = self
        while cv.parent is not None:
            to_return.append(cv.parent)
            cv = cv.parent
        return list(to_return)

    def descendants(self):
        # Using a recursive approach to obtain the full list of descendants
        if not self.child:
            return []
        else:
            # Getting the two lists of descendants of the two child and merging them in a single list
            child_descendants = sum([child.descendants() for child in self.child], [])
            # Returning the child plus their descendants
            return self.child[:] + child_descendants

    def create_child(self, _id_subset, _index):
        child = Job(id_reduced_precision=_id_subset,
                    forced_ids=self.forced_ids,
                    analysis_variables=self.analysis_variables,
                    banned_variables=self.banned_variables,
                    reduced_precision_level=self.reduced_precision_level,
                    communicator=self.communicator,
                    vault=self.vault,
                    template=self.template,
                    local_folder=self.local_folder,
                    result_filename=self.result_filename,
                    counter=self.counter
                    )

        child.parent = self
        child.child_number = _index
        child.level = self.level + 1
        return child

    def spawn_children(self):
        # Creates set for the two child
        _id_subset = self.divide_set_cluster()

        for index, id_subset in enumerate(_id_subset):
            child = self.create_child(id_subset, index)
            self.child.append(child)

    def run_children(self, _id_subset, _banned_variables, children_index):
        self.child = []
        for id_subset, banned_variables, child_index in (zip(_id_subset, _banned_variables, children_index)):
            child = self.create_child(id_subset, child_index)
            child.banned_variables = banned_variables
            child.reshuffle = True
            self.child.append(child)

    def divide_set_module(self, variables):
        import AutoRPE.UtilsRPE.Error as Error
        if len(variables) == 1:
            raise Error.ClusterCantBeDivided("This cluster can't be subdivided anymore")
        # Dividing by modules
        modules = [var.procedure.module.name for var in variables]
        # It is necessary to sort the modules otherwise the order can be different in different runs
        # leading to different groups with different hashes, which prevents reproducibility
        unique_modules = list(sorted(set(modules)))
        if len(unique_modules) > 1:
            d = {}
            for mod in unique_modules:
                d[mod] = modules.count(mod)
            d = {k: v for k, v in sorted(d.items(), key=lambda item: - item[1])}
            g1 = {"mod": [], "val": 0}
            g2 = {"mod": [], "val": 0}
            for key, value in d.items():
                if g1["val"] < g2["val"]:
                    g1["val"] += value
                    g1["mod"].append(key)
                else:
                    g2["val"] += value
                    g2["mod"].append(key)
            set1 = [v.id for v in variables if v.procedure.module.name in g1["mod"]]
            set2 = [v.id for v in variables if v.procedure.module.name in g2["mod"]]
            return set1, set2
        else:
            # Separate by routines
            routines = [var.procedure.name for var in variables]
            unique_routines = list(sorted(set(routines)))
            if len(unique_routines) > 1:
                d = {}
                for routine in unique_routines:
                    d[routine] = routines.count(routine)
                d = {k: v for k, v in sorted(d.items(), key=lambda item: - item[1])}
                g1 = {"mod": [], "val": 0}
                g2 = {"mod": [], "val": 0}
                for key, value in d.items():
                    if g1["val"] < g2["val"]:
                        g1["val"] += value
                        g1["mod"].append(key)
                    else:
                        g2["val"] += value
                        g2["mod"].append(key)
                set1 = [v.id for v in variables if v.procedure.name in g1["mod"]]
                set2 = [v.id for v in variables if v.procedure.name in g2["mod"]]
                return set1, set2
            else:
                # Just separate in two halves
                # assert len(self.id_reduced_precision) > 1
                set1 = []
                set2 = self.id_reduced_precision[:]
                for i in range(int(len(self.id_reduced_precision) / 2)):
                    set1.append(set2.pop(0))
                return set1, set2

    def divide_set_cluster(self):
        """Divide the set of variables to analyze taking into account """
        # Just one variable left, no possibility to further subdivide: fail
        if len(self.id_reduced_precision) == 1:
            raise Error.ClusterCantBeDivided("This cluster can't be subdivided anymore")

        # Divide the variables in the analysis set (not in the variable_set, where also forced can be found)
        variables = [self.vault.get_variable_by_id(_id) for _id in self.id_reduced_precision]

        # Get cluster ids
        cluster_id = [var.cluster_id for var in variables]
        # We are dealing with clusters
        if -1 not in cluster_id:
            # It is necessary to sort the clusters otherwise the order can be different in different runs
            # leading to different groups with different hashes, which prevents reproducibility
            cluster_id = list(set(cluster_id))
            cluster_id.sort()

            # Just one cluster: fail
            if len(cluster_id) == 1:
                raise Error.ClusterCantBeDivided("This cluster can't be subdivided anymore")
            # Two cluster left: divide into two branches
            elif len(cluster_id) == 2:
                set1 = [var.id for var in variables if var.cluster_id == cluster_id[0]]
                set2 = [var.id for var in variables if var.cluster_id == cluster_id[1]]
            # Various clusters: distribute the clusters on the branches in a balanced way
            else:
                set1 = []
                set2 = [var for var in variables if var.cluster_id != cluster_id[-1]]
                while len(cluster_id) > 0 and len(set1) < len(set2):
                    c_id = cluster_id.pop()
                    set1.extend([v for v in variables if v.cluster_id == c_id])
                    set2 = [v for v in set2 if v.cluster_id != c_id]
                set1 = [v.id for v in set1]
                set2 = [v.id for v in set2]
        # 4 - Just independent variables, either global or local: divide by module/subroutine
        else:
            set1, set2 = self.divide_set_module(variables)

        return set1, set2

    def divide_function_level(self, variables):
        sets = []
        levels = list(set([v.procedure.level for v in variables if v.procedure.name != 'main']))
        main_variables = [v.id for v in variables if v.procedure.name == 'main']
        if len(main_variables):
            sets.append(main_variables)
        if len(levels) == 1 or len(main_variables) == len(variables):
            return self.divide_set_module(variables)
        for l in levels:
            sets.append([v.id for v in variables if v.procedure.name != 'main' and v.procedure.level == l])
        return sets

    def kind_of_exception(self):
        """
        This method tries to categorize the kind of exception that occurs when two successful sets merge to an
        unsuccessful one.

        To start with something I propose:
        - No exception: It isn't a failed job
        - Intra-routine: All variables belong to same routine
        - Intra-module: All variables belong to same module, but not routine
        - Inter-module: The variables belong to different modules.
        :return: kind-of-exception
        """
        if not self.child:
            return "TreeLeafFail"
        child_1, child_2 = self.child
        # We'll focus only on the id of reduced precision
        ch1_variables = [self.vault.get_variable_by_id(_id) for _id in child_1.id_reduced_precision]
        ch2_variables = [self.vault.get_variable_by_id(_id) for _id in child_2.id_reduced_precision]
        ch1_routines = set([v.procedure for v in ch1_variables])
        ch2_routines = set([v.procedure for v in ch2_variables])
        if len(ch1_routines.union(ch2_routines)) == 1:
            return "IntraRoutine"
        ch1_modules = set([v.procedure.module for v in ch1_variables])
        ch2_modules = set([v.procedure.module for v in ch2_variables])
        if len(ch1_modules.union(ch2_modules)) == 1:
            return "IntraModule"
        return "InterModule"

    def graph(self):
        members = self.descendants()

        def style(status, incremental_id):
            if status == "SUCCESS":
                color = "#00b300"
            elif status == "FAILED":
                color = "#e60000"
            elif status == "SUSPENDED":
                color = "#ffff1a"
            else:
                color = "#aaa"
            return "style %s fill:%s\n" % (incremental_id, color)
            # return "style %s fill:%s\n" % (fix_identifier(node.identifier()), color)

        graph = "graph TD\n"
        graph += style(self.status, self.incremental_id)
        for member in members:
            len_id_reduced_precision = len(member.id_reduced_precision)
            len_banned_var = len(member.banned_variables)
            len_forced_var = len(member.forced_ids)
            ID = member.incremental_id

            if len_id_reduced_precision == 0:
                if len_banned_var:
                    node = "%s --> %s[ " + str(ID) + " : *+" + str(len_banned_var) + "]\n"
                else:
                    node = "%s --> %s[*]\n"
            else:
                if len_banned_var:
                    node = "%s --> %s[ " + str(ID) + " : " + str(len_id_reduced_precision) + "+" + str(len_banned_var) + "]\n"
                elif len_forced_var:
                    node = "%s --> %s[ " + str(ID) + " : " + str(len_id_reduced_precision) + "+ _" + str(len_forced_var) + "]\n"
                else:
                    node = "%s --> %s[" + str(ID) + " : " + str(len_id_reduced_precision) + "]\n"

            graph += node % (member.parent.incremental_id, member.incremental_id)

            graph += style(member.status, member.incremental_id)
        import calendar
        import time
        ts = calendar.timegm(time.gmtime())

        mermaid_graph_file = open(self.local_folder + "/mermaid/mermaid_chart_" + str(ts) + ".txt", "w")
        mermaid_graph_file.write(graph)
        mermaid_graph_file.close()
        return graph
