# Python structure to control the submission and status of a remote job with few other capabilities.


class RemoteManager:
    def __init__(self, id_reduced_precision,
                 forced_ids,
                 reduced_precision_level,
                 analysis_variables,
                 communicator,
                 vault,
                 template,
                 result_filename,
                 counter):
        self.vault = vault
        self.id_reduced_precision = id_reduced_precision
        self.forced_ids = [] if not forced_ids else forced_ids[:]
        self.analysis_variables = analysis_variables
        self.incremental_id = counter.up() if counter else 0
        self.communicator = communicator
        self.reduced_precision_level = reduced_precision_level
        self.job_name = "%i-%s" % (self.incremental_id, self.hash)
        self.num_proc = 48
        self.remote_scratch = self.communicator.remote_scratch
        self.remote_logdir = "%s/analysis/LOGS" % self.remote_scratch
        self.remote_namelist_dir = "%s/Namelists" % self.remote_logdir
        self.remote_rundir = "%s/analysis/%s" % (self.remote_scratch, self.hash)

        self.remote_jobscript_path = "%s/%s.cmd" % (self.remote_logdir, self.hash)
        self.remote_namelist = "%s/%s" % (self.remote_namelist_dir, self.hash)

        self.template = template
        self.result_filename = result_filename

        self.jobscript = self.generate_jobscript()

        self.elapsed_time = None
        self.job_id = None
        self._result = None

        # The default status is PENDING
        self._status = "PENDING"

    @property
    def variable_set(self):
        if self.forced_ids:
            return self.id_reduced_precision + self.forced_ids
        else:
            return self.id_reduced_precision

    @property
    def hash(self):
        """
            Given a list of integers corresponding to variable ids, it generates a unique hash that identifies it
        """
        import hashlib
        # Sort in order of raising id
        self.variable_set.sort()
        # Convert to string the list of ids
        string_to_hash = "".join([str(i_id) + '_' + str(self.reduced_precision_level) for i_id in self.variable_set])
        # Create the hash
        hash_object = hashlib.md5(string_to_hash.encode())
        return hash_object.hexdigest()

    def variables(self):
        return [self.vault.get_variable_by_id(_x) for _x in set(self.variable_set)]

    @property
    def remote_status(self):
        _status, _time = self.job_status()
        _status = _status.strip()
        return _status

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, new_value):
        possible_status = ["PENDING", "RUNNING", "ASSERTION_PENDING", "SUCCESS", "FAILED", "SUSPENDED"]
        # The list of possible status shall cover all the possible situations in which a job can be found.
        # PENDING: The job has been created and ready to be launched.
        # A PENDING JOB CAN ONLY HAVE SUSPENDED ANCESTORS AND SUCCESS OR FAILED DESCENDANTS
        # RUNNING: The job is RUNNING.
        # A RUNNING JOB CAN ONLY HAVE SUSPENDED ANCESTORS AND SUCCESS OR FAILED DESCENDANTS.
        # SUCCESS: A job that has been completed with successful results and asserted.
        # A SUCCESS job can not have both descendants FAILED
        # FAILED: A job that has been completed with unsuccessful results and that can not be
        # split ( Already was or a single variable set).
        # SUSPENDED: A job that has been completed with unsuccessful results that has been split and depends on the
        # results of descendant jobs.
        # SUSPENDED ancestors must be SUSPENDED.
        # At least one direct descendant should be different from FAILED or COMPLETED
        if new_value in possible_status:
            self._status = new_value
        else:
            raise AssertionError("Unknown status")

    def generate_jobscript(self):
        """
        Replaces a list of parameters from a provided dictionary in a job template.
        Inputs:
             - template (str): A text job with has some parameters between %% (i.e. %USERNAME%) which will be replaced.
             - parameters (dictionary): A dictionary including keys and values.
        Returns:
             - job (str): text corresponding to a job that can be submitted to the compute nodes.
        """
        with open(self.template) as f:
            job = f.read()
            job = job.replace("%REMOTE_PATH%", self.remote_scratch)
            job = job.replace("%RUNHASH%", self.hash)
            job = job.replace("%JOBNAME%", self.job_name)
            job = job.replace("%LOGDIR%", self.remote_logdir)
            job = job.replace("%NAMELIST%", self.remote_namelist)
        return job

    def check_if_result_already_exists(self):
        self.update_parameters()
        job_id_filename = "%s/job_id.txt" % self.remote_rundir
        try:
            # Recovering job_id to later check its status
            self.job_id = self.communicator.sftp.open(job_id_filename).read().decode().strip()
        except IOError:
            return False
        return True

    def check_if_already_submitted(self):
        self.update_parameters()
        try:
            command = 'squeue --format="%i %j"'
            stdin, stdout, stderr = self.communicator.execute(command)

            output = stdout.read().decode()
            lines = output.split("\n")
            if len(lines) > 1:
                jobs_in_remote_queue = lines[1:]
                for line in jobs_in_remote_queue:
                    if line.strip():
                        job_id, job_name = line.split()
                        if job_name == self.job_name:
                            return job_id

        except IOError:
            return None
        return None

    def generate_namelist(self):
        """
        It takes a list of variable ids and generates a RPE namelist which sets the variables's precision to
        the reduced precision level provided as argument.

        """
        # Namelist blueprint
        line_text = "emulator_variable_precisions(%d) = %d\t! Variable:%24s\tRoutine:%22s\tModule:%16s\n"
        # Namelist text
        namelist = ["! namelist variable precisions\n&precisions\n"]
        # Order variables by id ascending order
        self.analysis_variables.sort(key=lambda x: x.id)
        # Fill namelist
        for v in self.analysis_variables:
            # Reduce precision if variable is being analyzed otherwise keep original
            v_precision =  self.reduced_precision_level if v.id in self.variable_set else v.precision
            namelist.append(
                line_text % (v.id, v_precision, v.name, v.procedure.name, v.procedure.module.name))

        namelist.append("/\n")
        return "".join(namelist)

    def run_job(self, force=False):
        # Submit the job if force is true or the results are not present yet
        if force or not self.check_if_result_already_exists():
            # Generate namelist

            # Check if the job is already in the queue
            job_id = self.check_if_already_submitted()
            if job_id:
                # Set the got from the remote
                self.job_id = job_id
                print("Job %s was already in queue with job_id %s" % (self.incremental_id, self.job_id))
            else:
                # The job has not been submitted yet
                self.job_id = self.submit_job()
                print("Job %s successfully submitted with job_id %s" % (self.incremental_id, self.job_id))

        else:
            return

    def submit_job(self):
        """
        Saves a job (text) to a file in the remote machine and submits it to the scheduler.
        """
        import re
        # Write the namelist to remote file
        self.communicator.write_file(self.generate_namelist(), self.remote_namelist)
        # Write job to remote file
        self.communicator.write_file(self.jobscript, self.remote_jobscript_path)
        # with self.communicator.sftp.file(self.remote_jobscript_path, "w") as jobscript:
        #     jobscript.write(self.jobscript)

        # Defining the scheduler submit command
        # TODO: To use this tool with other schedulers, it might be convenient to define it somewhere else.
        scheduler_submit_command = "sbatch"

        # Defining the command to submit the job
        command = "%s %s" % (scheduler_submit_command, self.remote_jobscript_path)

        # Executing the command using the communicator
        stdin, stdout, stderr = self.communicator.execute(command)
        output = str(stdout.read())

        # Read the job id of the submitted job
        job_id_pattern = "Submitted batch job +([0-9]+)"
        job_id = re.search(job_id_pattern, output).group(1)
        # Return job id
        return job_id

    def job_status(self):
        """
        Given a job id, it returns the status of the job in the remote machine.
        """
        # Defining the scheduler status command
        # TODO: To use this tool with other schedulers, it might be convenient to define it somewhere else.
        scheduler_status_command = 'sacct  --format="State,Elapsed"'

        # Defining the command that will be sent to the remote machine
        command = '%s --job %s' % (scheduler_status_command, self.job_id)

        # Executing the command using the communicator
        stdin, stdout, stderr = self.communicator.execute(command)

        # Read status and time in the output received.
        output = stdout.read().decode()
        try:
            split_output = output.split("\n")[2].split()
            status = split_output[0]
            time = split_output[1]
        except IndexError:
            status = "PENDING"
            time = "00:00"
        # Return status and time
        return status, time

    def update_parameters(self):
        from os.path import isdir
        from os import mkdir
        job_hash = self.hash
        self.remote_rundir = "%s/analysis/%s" % (self.remote_scratch, job_hash)
        self.job_name = "%i-%s" % (self.incremental_id, job_hash)

        self.remote_jobscript_path = "%s/%s.cmd" % (self.remote_logdir, job_hash)
        self.remote_namelist = "%s/%s" % (self.remote_namelist_dir, job_hash)

        self.jobscript = self.generate_jobscript()
